<?php

/**
 * Copyright (c) 2014, MChart author(s)
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 *  purpose with or without fee is hereby granted, provided that the above
 *  copyright notice and this permission notice appear in all copies.
 */

namespace mchart;

abstract class IChart {
	
	protected $series = [];
	protected $colors = [];
	protected $legend = [];
	
	public $backgroundcolor = 0xFFFFFF;
	public $bordercolor     = 0x000000;
	public $xaxisrenderer   = null;
	public $yaxisrenderer   = null;
	
	/**
	 * Construct new \IChart class.
	 */
	public function __construct() {
		$this->xaxisrenderer = function($dp) { return ''; };
		$this->yaxisrenderer = function($dp) { return ''; };		
	}
	
	/**
	 * Add a data series to render in this graph.
	 * 
	 * @param array $data     Array of [numeric x, numeric y] data tuples
	 * @param string $legend  Text string to display in legend
	 * @param int $color      Color in hex literal format e.g. 0xFF0000
	 */
	public function addSeries(array $data, $legend=null, $color=null) {
		$this->series[] = $data;
		
		if (is_null($legend)) {
			$legend = 'Series '.(count($this->series));
		}
		$this->legend[] = $legend;
		
		if (is_null($color)) {
			$defaultColors = [
				0x808080, 0xFF0000, 0x00FF00, 0x0000FF,
				0xFF00FF, 0x00FFFF, 0xFFFF00
			];
			$color = $defaultColors[ count($this->series) % count($defaultColors) ];
		}		
		$this->colors[] = $color;
	}
	
	/**
	 * Create a GD brush handle for an RGB color.
	 * 
	 * @param resource $img GD handle
	 * @param int $rgb
	 * @return resource GD brush handle
	 */
	protected function colorcreatergb($img, $rgb) {
		$b = $rgb & 0xFF;
		$g = ($rgb >> 8) & 0xFF;
		$r = ($rgb >> 16) & 0xFF;
		
		return imagecolorallocate($img, $r, $g, $b);
	}
	
	/**
	 * Retrieve dimensions of a GD font.
	 * 
	 * @param int $font GD font (1..5 inclusive)
	 * @return array An array [int width, int height]
	 */
	public static function getGDFontDimensions($font) {
		$width  = [0, 5, 6, 7,  8,  9];
		$height = [0, 6, 8, 13, 15, 15];
		return [$width[$font], $height[$font]];
	}
		
	/**
	 * Render all series to PNG image.
	 * 
	 * @param int $width   Output image width in pixels
	 * @param int $height  Output image height in pixels
	 * @param int $font    GD font to use for labels and legends (1..5 inclusive)
	 * @return string      Raw PNG data
	 */
	public function render($width, $height, $font=1) {
		$xmax = 1;
		$ymax = 1;
		$ymin = 0;
		list($fontwidth, $fontheight) = self::getGDFontDimensions($font);
		$xaxispadding = 35;
		$yaxispadding = 20;
				
		$img         = imagecreate($width, $height);		
		$bgbrush     = $this->colorcreatergb($img, $this->backgroundcolor);
		$borderbrush = $this->colorcreatergb($img, $this->bordercolor);
		
		$array_select = function(array $a, $key) {
			$ret = [];
			foreach($a as $k => $r) { $ret[$k] = $r[$key]; }
			return $ret;
		};
		
		// Find maximum X and Y points to determin scale factors
		foreach($this->series as $seriesdata) {
			$xmax = max([$xmax, max($array_select($seriesdata, 0))]);
			$ymax = max([$ymax, max($array_select($seriesdata, 1))]);
			$ymin = min([$ymin, min($array_select($seriesdata, 1))]);
		}
		
		$ytransform = function($y) use ($height, $yaxispadding, $ymax, $ymin) {			
			$yscale = ($height - $yaxispadding) / (abs($ymax) + abs($ymin));			
			$ypos   = $height - (($y-$ymin) * $yscale) - $yaxispadding;
			
			return $ypos;
		};
		
		$xscale = ($width  - $xaxispadding) / $xmax;
		$xtransform = function($x) use ($xscale, $xaxispadding) {
			return $x * $xscale + $xaxispadding;
		};
		
		// Draw background
		imagerectangle($img, 0, 0, $width, $height, $bgbrush);
		
		// Plot data points
		$this->renderDataPoints($img, $xtransform, $ytransform);
		
		// Draw X, Y axis		
		imageline($img, $xaxispadding, 0, $xaxispadding, $height - $yaxispadding, $borderbrush);
		imageline($img, $xaxispadding, $ytransform(0), $width, $ytransform(0), $borderbrush);
		
		// Axis labels
		for($i = 0; $i < $xmax; $i++) {
			$write = call_user_func($this->xaxisrenderer, $i);
			if (strlen($write)) {
				$xmark = ($i * $xscale + $xaxispadding);
				$ymark = $ytransform(0);
				$yoff  = $ymark + ($fontheight/2);
				imageline($img, $xmark, $ymark - 2, $xmark, $ymark + 2, $borderbrush); // marker
				imagestring($img, $font, $xmark - (strlen($write) * $fontwidth)/2, $yoff, $write, $borderbrush); // text
			}
		}
		
		for ($i = $ymin; $i < $ymax; $i++) {
			$write = call_user_func($this->yaxisrenderer, $i);
			if (strlen($write)) {
				$textwidth = $fontwidth * strlen($write);
				$ymark = $ytransform($i);
				imageline($img, $xaxispadding-2, $ymark, $xaxispadding+2, $ymark, $borderbrush); // marker
				imagestring($img, $font, $xaxispadding-$textwidth-2, $ymark-4, $write, $borderbrush);
			}
		}
		
		// Draw legend		
		foreach($this->legend as $i => $label) {
			$brush = $this->colorcreatergb($img, $this->colors[$i]);
			
			$ymark = 15 * $i;
			$liney = $ymark + ($fontheight / 2);
			
			imageline($img, $xaxispadding + 2, $liney,   $xaxispadding + 2 + 10, $liney,   $brush);
			imageline($img, $xaxispadding + 2, $liney+1, $xaxispadding + 2 + 10, $liney+1, $brush);
			imagestring($img, $font, $xaxispadding + 2 + 10 + 2, $ymark, $label, $brush);
		}
		
		// Rasterise PNG
		ob_start();
		imagepng($img);
		return ob_get_clean();
	}
		
	protected abstract function renderDataPoints($img, $xtransform, $ytransform);
}

class LineChart extends IChart {	
	
	protected function renderDataPoints($img, $xtf, $ytf) {
		foreach($this->series as $sx => $seriesdata) {
			$brush = $this->colorcreatergb($img, $this->colors[$sx]);
			
			for ($i = 1, $e = count($seriesdata); $i !== $e; ++$i) {
				list($x, $y) = $seriesdata[$i];
				list($oldx, $oldy) = $seriesdata[$i-1];
				
				imageline($img, $xtf($oldx), $ytf($oldy), $xtf($x), $ytf($y), $brush);
			}
		}
	}
	
}

class StaggeredLineChart extends IChart {
	
	protected function renderDataPoints($img, $xtf, $ytf) {
		foreach($this->series as $sx => $seriesdata) {
			$brush = $this->colorcreatergb($img, $this->colors[$sx]);
			
			for ($i = 1, $e = count($seriesdata); $i !== $e; ++$i) {
				list($x, $y) = $seriesdata[$i];
				list($oldx, $oldy) = $seriesdata[$i-1];
				
				imageline($img, $xtf($oldx), $ytf($oldy), $xtf($x), $ytf($oldy), $brush);
				imageline($img, $xtf($x), $ytf($oldy), $xtf($x), $ytf($y), $brush);
				
			}
		}
	}
	
}

class BarChart extends IChart {
	
	protected function renderDataPoints($img, $xtransform, $ytransform) {
		
		$y0 = $ytransform(0);
		
		foreach($this->series as $sx => $seriesdata) {
			$brush = $this->colorcreatergb($img, $this->colors[$sx]);
			
			for ($i = 0, $e = count($seriesdata); $i !== $e; ++$i) {
				list($x, $y) = $seriesdata[$i];
				
				$x1 = $xtransform($x + ( $sx   / count($this->series)));
				$x2 = $xtransform($x + (($sx+1)/ count($this->series)));
				
				imagefilledrectangle($img, $x1, $y0, $x2, $ytransform($y), $brush);
			}
		}
	}	
}
